#include "cubeMap.h"
#include "material.h"
#include "RTConfig.h"
#include <qmessagebox.h>

// Pour ne pas avoir a mettre qglviewer:: devant les classes
// definies dans le namespace qglviewer.
using namespace qglviewer;

CubeMap::CubeMap():bradius_(1.0f) {
}

void CubeMap::draw() const
{
  
}

void CubeMap::initFromDOMElement(const QDomElement& e)
{
 
	if (e.tagName() != "CubeMap") {
      QMessageBox::critical(NULL, "CubeMap init error", "CubeMap::initFromDOMElement, bad DOM tagName.\nExpecting 'CubeMap', got "+e.tagName());
      return;
    }

  //Object::initFromDOMElement(e);

  QDomNode n = e.firstChild();

  while (!n.isNull()) {
      QDomElement e = n.toElement();
      if (!e.isNull()) {
		/*if(e.tagName() == "Scale")
			setTextureScale(e.attribute("scale", "1.0").toFloat());*/

		 if (e.tagName() == "MapPosX")
			  loadMap(e.attribute("file"), 0);
		 else if (e.tagName() == "MapNegX")
			  loadMap(e.attribute("file"), 1);
		 else if (e.tagName() == "MapPosY")
			  loadMap(e.attribute("file"), 2);
		 else if (e.tagName() == "MapNegY")
			  loadMap(e.attribute("file"), 3);
		 else if (e.tagName() == "MapPosZ")
			  loadMap(e.attribute("file"), 4);
		 else if (e.tagName() == "MapNegZ")
			 loadMap(e.attribute("file"), 5);

		 else
			  QMessageBox::warning(NULL, "CubeMap child error", "Unsupported CubeMap  child : "+e.tagName()); 
	  }
      else
		  QMessageBox::warning(NULL, "CubeMap  XML error", "Error while parsing CubeMap  XML document");

      n = n.nextSibling();
    }
}

void CubeMap::loadMap(const QString& fileName, int n){
  n=n>0 ? n%6 : 0;

  if (fileName.isNull())
    QMessageBox::warning(NULL, "CubeMap  texture error", "CubeMap error : no texture file name provided");
  else
    if (!this->cubeImages_[n].load(fileName))
      QMessageBox::warning(NULL, "CubeMap  texture error", "Unable to load CubeMap texture from "+fileName);
}

Color CubeMap::getMapColor(float u, float v, int side) const{

  // Take texture scaling into account
/*	u /= (float)this->cubeImages_[side].width();
	v /= (float)this->cubeImages_[side].height();*/

  // Compute pixel coordinates
  const int i = int(fabs(u - floor(u)) * this->cubeImages_[side].width());
  const int j = int(fabs(v - floor(v)) * this->cubeImages_[side].height());

  return Color(this->cubeImages_[side].pixel(i,j));
}

bool CubeMap::intersect(const Ray& ray, Hit& hit) const
{

	/*if(hit.directColor)
		return false;*/

    Color outputColor = Color(0.0f,0.0f,0.0f);
	qglviewer::Vec raydir;

	raydir.x=ray.direction().x;
	raydir.y=ray.direction().z;
	raydir.z=ray.direction().y;

    
	if ((fabsf(raydir.x) >= fabsf(raydir.y)) && (fabsf(raydir.x) >= fabsf(raydir.z)))
    {
        if (raydir.x > 0.0f){
			outputColor=getMapColor(1.0f - (raydir.z / raydir.x+ 1.0f) * 0.5f, 
									(raydir.y / raydir.x+ 1.0f) * 0.5f, CM_POS_X);
        }
        else if (raydir.x < 0.0f)
        {
			outputColor=getMapColor(1.0f - (raydir.z / raydir.x+ 1.0f) * 0.5f, 
									1.0f - (raydir.y / raydir.x+ 1.0f) * 0.5f, CM_NEG_X);

            /*currentColor = cm.texture + cubemap::left * cm.sizeX * cm.sizeY;
            outputColor = readTexture(currentColor, 
                1.0f - (myRay.dir.z / myRay.dir.x+ 1.0f) * 0.5f,
                1.0f - ( myRay.dir.y / myRay.dir.x + 1.0f) * 0.5f, 
                cm.sizeX, cm.sizeY);*/
        }
    }
    else if ((fabsf(raydir.y) >= fabsf(raydir.x)) && (fabsf(raydir.y) >= fabsf(raydir.z)))
    {
        if (raydir.y > 0.0f)
        {
			outputColor=getMapColor((raydir.x / raydir.y+ 1.0f) * 0.5f, 
									1.0f - (raydir.z / raydir.y+ 1.0f) * 0.5f, CM_POS_Y);
        }
        else if (raydir.y < 0.0f)
        {
			outputColor=getMapColor(1.0f - (raydir.x / raydir.y+ 1.0f) * 0.5f, 
									1.0f-(raydir.z / raydir.y+ 1.0f) * 0.5f, CM_NEG_Y);

            /*currentColor = cm.texture + cubemap::down * cm.sizeX * cm.sizeY;
            outputColor = readTexture(currentColor, 
                1.0f - (myRay.dir.x / myRay.dir.y + 1.0f) * 0.5f, 
                (myRay.dir.z/myRay.dir.y + 1.0f) * 0.5f, cm.sizeX, cm.sizeY);*/
        }
    }
    else if ((fabsf(raydir.z) >= fabsf(raydir.x)) && (fabsf(raydir.z) >= fabsf(raydir.y)))
    {
        if (raydir.z > 0.0f)
        {
			outputColor=getMapColor((raydir.x / raydir.z+ 1.0f) * 0.5f, 
									(raydir.y / raydir.z+ 1.0f) * 0.5f, CM_POS_Z);


           /* currentColor = cm.texture + cubemap::forward * cm.sizeX * cm.sizeY;
            outputColor = readTexture(currentColor, 
                (myRay.dir.x / myRay.dir.z + 1.0f) * 0.5f, 
                (myRay.dir.y/myRay.dir.z + 1.0f) * 0.5f, cm.sizeX, cm.sizeY);*/
        }
        else if (raydir.z < 0.0f)
        {
			outputColor=getMapColor((raydir.x / raydir.z+ 1.0f) * 0.5f, 
									1.0f - (raydir.y / raydir.z+ 1.0f) * 0.5f, CM_NEG_Z);


            /*currentColor = cm.texture + cubemap::backward * cm.sizeX * cm.sizeY;
            outputColor = readTexture(currentColor, 
                (myRay.dir.x / myRay.dir.z + 1.0f) * 0.5f, 
                1.0f - (myRay.dir.y /myRay.dir.z+1) * 0.5f, cm.sizeX, cm.sizeY);*/
        }
    }


	hit.color_=outputColor;
	hit.directColor=true;
	

    return false;
}
