#include "sphere.h"
#include "material.h"
#include "RTConfig.h"

// Pour ne pas avoir a mettre qglviewer:: devant les classes
// definies dans le namespace qglviewer.
using namespace qglviewer;

#define GL_BGRA 0x80E1

Sphere::Sphere():slices(10), stacks(10), imgTexID_(0) {
  setRadius(1.0);
}

void Sphere::draw() const
{

	if(!RTConfig::getInstance()->drawSphere)
		return;

	//Disable rendering for gpurt safe mode
	/*if(RTConfig::getInstance()->gpurtSafeMode)
		return;*/

  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();

  glPushAttrib(GL_ALL_ATTRIB_BITS);

  glMultMatrixd(frame().matrix());
  glColor3fv(material().diffuseColor());


//???????????????????????????
//->Ca c'est pour l'eclairage OpenGL, pour faire du blinn-phong :)
  float colorMat[] = { material().diffuseColor().r, material().diffuseColor().g, material().diffuseColor().b, 1.0f };
  glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, colorMat);
  float colorMatSpec[] = { material().specularColor().r, material().specularColor().g, material().specularColor().b, 1.0f };
  glMaterialfv(GL_FRONT, GL_SPECULAR, colorMatSpec);
  glMateriali(GL_FRONT, GL_SHININESS, (int)material().specularCoefficient());
//???????????????????????????

/**/    

  glEnable(GL_DEPTH_TEST);
  //glDisable(GL_LIGHTING);
  

  if(!this->material().texture().isNull()){
	  if(!this->imgTexID_) {
		glGenTextures(1, &(this->imgTexID_));

		glBindTexture(GL_TEXTURE_2D, this->imgTexID_);

		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S, GL_REPEAT);
		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T, GL_REPEAT);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA8, this->material().texture().width(), this->material().texture().height(), 0, GL_BGRA, GL_UNSIGNED_BYTE, this->material().texture().bits() );

		//std::cout<<this->material().texture().width()<<" "<<this->material().texture().height()<<"\n";
	  }

	  glEnable(GL_TEXTURE_2D);
	  glBindTexture(GL_TEXTURE_2D, this->imgTexID_);

  }else{
	  glDisable(GL_TEXTURE_2D);
  }

  GLUquadricObj *quad=gluNewQuadric();
  gluQuadricNormals(quad, GLU_SMOOTH);
  gluQuadricTexture(quad, GL_TRUE);
  
  glMatrixMode(GL_TEXTURE);
  glPushMatrix();
  glScalef(1,-1,1);

  glTranslatef(-0.25,0,0);

  glScalef(1.0f/this->material().textureScaleU(), 1.0f/this->material().textureScaleV(), 1);

  /*if(this->material().textureMode()==Material::BLEND)
	  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_BLEND);*/


  glDisable(GL_BLEND);

  gluSphere(quad,this->radius() , slices, stacks);
//  glBindTexture(GL_TEXTURE_2D, 0);
  glDisable(GL_TEXTURE_2D);
 
  gluDeleteQuadric(quad);

  glMatrixMode(GL_TEXTURE);
  glPopMatrix();

  glMatrixMode(GL_MODELVIEW);
  glPopAttrib();
  glPopMatrix();
}

// Build a sphere object
void Sphere::initFromDOMElement(const QDomElement& e)
{
  Object::initFromDOMElement(e);
  if(e.hasAttribute("radius"))
    this->setRadius(e.attribute("radius", "0").toFloat());
}


bool Sphere::intersect(const Ray& ray, Hit& hit) const
{
  // In the frame of the sphere (center = (0,0,0) radius = r), 
  // with Ro the origin of the ray, and Rd its direction,
  // we have to solve the equation: Rd*Rd*t + 2*Rd*Ro*t + Ro*Ro - r = 0
  //                                   a *t +    b   *t +      c     = 0
  // BUT here, we are in the world frame. We must replace us in the frmae of the sphere
  // SO:
  //    Rd = ray.direction()  --> unchanged  
  //    Ro = ray.start()-this->frame().position()

  float a=1;  // Rd is normalized
  float b=2.0f*(ray.direction())*(ray.start()-this->frame().position());
  float c=(ray.start()-this->frame().position())*(ray.start()-this->frame().position())-this->radius()*this->radius();

  // Compute the discriminant
  float d=(b*b-4.0f*a*c);
  float t;

  // If d<0, no solution
  if(d<0){
	  return false;
  }

  // To avoid noise if the ray intersects the object itself;
  float tmin=RTConfig::getInstance()->getRayTMin();// 0.0001;

  // If d==0, one solution (ie the ray intersects "on" the tangent of the sphere: nearly impossible with float precision!)
  if(d==0)
	t=-b/(2*a);
  
  // If d>0, 2 solutions (front and back of the sphere)
  else{
	float t0=(-b-sqrt(d))/(2.0f*a);
	float t1=(-b+sqrt(d))/(2.0f*a);

	// Gets the minimum t (ie the first intersection from the origin of the ray and the object)
	//t=(t0<t1)?t0:t1;

	if( t0<tmin && t1>tmin)
		t=t1;
	else if(t1<tmin && t0>tmin)
		t=t0;
	else
		t=(t0<t1)?t0:t1;

  }


  // If the "new computed t" is < to the hit.time, we keep it (this intersection is the nearest)
  // (we must have t>tmin in order not to be behind the origin of the ray)
  if(hit.time()>t && t>tmin){
	  // Modify all the parameters of the hit
	  hit.setTime(t);
	  hit.setIntersection( ray.direction()*t+ray.start() );
	  hit.setMaterial(this->material());

	  // Compute the normale at the intersection (ie direction of the radius from the center of the sphere)
	  Vec normale = hit.intersection()-this->frame().position();
	  normale.normalize();
	  hit.setNormal(normale);
	  
	  // "Partie XI: Textures"
	  qglviewer::Vec pos = hit.intersection() - this->frame().position();
	  pos=this->frame().rotation().rotate(pos);
	  hit.setUV((atan2(pos.y,pos.x)+M_PI)/(2.0f*M_PI), (acosf(pos.z/this->radius()))/(M_PI) );
    

	  ///Tangent space computation
	  // Extension: Bump Mapping
	  Vec tangent =  Vec(1,0,0);
	  Vec bitangent =  Vec(0,1,0);
	 
	  Vec tangentTmp, bitangentTmp;

	  float u = hit.u() * (2.0f*M_PI) + M_PI;
	  float v = hit.v() * M_PI;

	  // Rotation along z-axis
	  tangentTmp.x = cos(u) * tangent.x - sin(u) * tangent.y;
	  tangentTmp.y = sin(u) * tangent.x + cos(u) * tangent.y;
	  tangentTmp.z = 0; //perturbNormale.z
	  // Rotation along x-axis
	  tangent.x = tangentTmp.x;
	  tangent.y = cos(v) * tangentTmp.y - sin(v) * tangentTmp.z;
	  tangent.z = sin(v) * tangentTmp.y + cos(v) * tangentTmp.z;

	  // Rotation along z-axis
	  bitangentTmp.x = cos(u) * bitangent.x - sin(u) * bitangent.y;
	  bitangentTmp.y = sin(u) * bitangent.x + cos(u) * bitangent.y;
	  bitangentTmp.z = 0; //perturbNormale.z
	  // Rotation along x-axis
	  bitangent.x = tangentTmp.x;
	  bitangent.y = cos(v) * bitangentTmp.y - sin(v) * bitangentTmp.z;
	  bitangent.z = sin(v) * bitangentTmp.y + cos(v) * bitangentTmp.z;
	  

	  hit.setTangentSpace(tangent, bitangent);
	


	 /* // Extension: Bump Mapping
	  Vec perturbNormale =  hit.material().bumpMapColor(hit.u(), hit.v());
	  Vec rotatePerturbNormale;
	
	  if ( (perturbNormale.x == -1) && (perturbNormale.y == -1) && (perturbNormale.z == -1) ) 
		  rotatePerturbNormale = Vec(0,0,0);//hit.normal();
	  else {
		  Vec rotatePerturbNormaleTmp;
		  float u = hit.u() * (2.0f*M_PI) + M_PI;
		  float v = hit.v() * M_PI;

		  // Rotation along z-axis
		  rotatePerturbNormaleTmp.x = cos(u) * perturbNormale.x - sin(u) * perturbNormale.y;
		  rotatePerturbNormaleTmp.y = sin(u) * perturbNormale.x + cos(u) * perturbNormale.y;
		  rotatePerturbNormaleTmp.z = 0; //perturbNormale.z
		  // Rotation along x-axis
		  rotatePerturbNormale.x = rotatePerturbNormaleTmp.x;
		  rotatePerturbNormale.y = cos(v) * rotatePerturbNormaleTmp.y - sin(v) * rotatePerturbNormaleTmp.z;
		  rotatePerturbNormale.z = sin(v) * rotatePerturbNormaleTmp.y + cos(v) * rotatePerturbNormaleTmp.z;
	  }

	  Vec newNormale = hit.normal() + rotatePerturbNormale/1.0f;
	  newNormale.normalize();


	  hit.setbumpMapNormal(newNormale);
	  hit.setNormal(newNormale);*/


	//  hit.setCoeffSpecularMap(hit.material().specularMapColor(hit.u(), hit.v()));

	  return true;
  }

  return false;
}
